local addon, private = ...
local mod = LibStub("AceAddon-3.0"):NewAddon(addon, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local ArtifactUI = _G.C_ArtifactUI

function getSocketSlotID()
  local class = select(2, UnitClass("player"))
  if (class == "PALADIN" and GetSpecialization() == 2) or (class == "WARRIOR" and GetSpecialization() == 3) then
    return 17
  else
    return 16
  end
end

function mod:Update()
  self:BuildOutput()
  self.CopyFrame:Show()
end

function mod:copySlotsFromEquipped()
  for _, frame in ipairs(self.frameSlots) do
    local item = nil
    if frame.info.slot_id and not frame.info.artifact_gem_id then
      item = GetInventoryItemLink('player', frame.info.slot_id)
    elseif frame.info.artifact_gem_id then
      local _, gemLink = GetItemGem(GetInventoryItemLink('player', frame.info.slot_id), frame.info.artifact_gem_id)
      item = gemLink
    end
    frame.setItem(item, item)
  end
  self:Update()
end

function mod:OnInitialize()
  mod:RegisterChatCommand('sim', self.loadWindow)
end

function mod:BuildTalentString()
  local talentInfo = {}
  local maxTiers = 7
  local maxColumns = 3
  for tier = 1, maxTiers do
    for column = 1, maxColumns do
      local talentID, name, iconTexture, selected, available = GetTalentInfo(tier, column, GetActiveSpecGroup())
      if selected then
        talentInfo[tier] = column
      end
    end
  end

  local str = 'talents='
  for i = 1, maxTiers do
    if talentInfo[i] then
      str = str .. talentInfo[i]
    else
      str = str .. '0'
    end
  end

  return str
end

local lastArtifactString = nil
function mod:BuildArtifactString()
  local item_id = select(1, ArtifactUI.GetArtifactInfo())
  if item_id == nil or item_id == 0 then
    return lastArtifactString
  end

  local artifact_id = self.ARTIFACT_TABLE[item_id]
  if artifact_id == nil then
    return lastArtifactString
  end

  -- Note, relics are handled by the item string
  local str = 'artifact=' .. artifact_id .. ':0:0:0:0'

  local powers = ArtifactUI.GetPowers()
  for i = 1, #powers do
    local power_id = powers[i]
    local _, _, currentRank, _, bonusRanks = ArtifactUI.GetPowerInfo(power_id)
    if currentRank > 0 and currentRank - bonusRanks > 0 then
      str = str .. ':' .. power_id .. ':' .. (currentRank - bonusRanks)
    end
  end

  lastArtifactString = str
  return str
end

function mod:BuildSimcLine(itemLink, useEquippedRelics)
  local itemName = GetItemInfo(itemLink)
  local itemString = string.match(itemLink, "item:([%-?%d:]+)")
  local itemSplit = {}
  local simcItemOptions = {}
  local slug_name = gsub(gsub(string.lower(itemName), " ", "_"), "[^%a_]", "")
  tinsert(simcItemOptions, slug_name)

  -- Split data into a table
  for v in string.gmatch(itemString, "(%d*:?)") do
    if v == ":" then
      itemSplit[#itemSplit + 1] = 0
    else
      itemSplit[#itemSplit + 1] = string.gsub(v, ':', '')
    end
  end

  -- Item id
  local itemId = itemSplit[self.OFFSETS.ITEM_ID]
  simcItemOptions[#simcItemOptions + 1] = 'id=' .. itemId

    -- Enchant
  if tonumber(itemSplit[self.OFFSETS.ENCHANT_ID]) > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'enchant_id=' .. itemSplit[self.OFFSETS.ENCHANT_ID]
  end

  -- New style item suffix, old suffix style not supported
  if tonumber(itemSplit[self.OFFSETS.SUFFIX_ID]) ~= 0 then
    simcItemOptions[#simcItemOptions + 1] = 'suffix=' .. itemSplit[self.OFFSETS.SUFFIX_ID]
  end

  local flags = tonumber(itemSplit[self.OFFSETS.FLAGS])
  local bonuses = {}

  for index=1, tonumber(itemSplit[self.OFFSETS.BONUS_ID]) do
    bonuses[#bonuses + 1] = itemSplit[self.OFFSETS.BONUS_ID + index]
  end

  if #bonuses > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'bonus_id=' .. table.concat(bonuses, '/')
  end

  local rest_offset = self.OFFSETS.BONUS_ID + #bonuses + 1

  -- Upgrade level
  if bit.band(flags, 4) == 4 then
    local upgrade_id = tonumber(itemSplit[rest_offset])
    if self.UPGRADE_TABLE[upgrade_id] ~= nil and self.UPGRADE_TABLE[upgrade_id] > 0 then
      simcItemOptions[#simcItemOptions + 1] = 'upgrade=' .. self.UPGRADE_TABLE[upgrade_id]
    end
    rest_offset = rest_offset + 1
  end

  -- Artifacts use this
  if bit.band(flags, 256) == 256 then
    rest_offset = rest_offset + 1 -- An unknown field
    local relic_str = ''
    while rest_offset < #itemSplit do
      local n_bonus_ids = tonumber(itemSplit[rest_offset])
      rest_offset = rest_offset + 1

      if n_bonus_ids == 0 then
        relic_str = relic_str .. 0
      else
        for rbid = 1, n_bonus_ids do
          relic_str = relic_str .. itemSplit[rest_offset]
          if rbid < n_bonus_ids then
            relic_str = relic_str .. ':'
          end
          rest_offset = rest_offset + 1
        end
      end

      if rest_offset < #itemSplit then
        relic_str = relic_str .. '/'
      end
    end

    if relic_str ~= '' then
      simcItemOptions[#simcItemOptions + 1] = 'relic_id=' .. relic_str
    end
  end

  -- Some leveling quest items seem to use this, it'll include the drop level of the item
  if bit.band(flags, 512) == 512 then
    simcItemOptions[#simcItemOptions + 1] = 'drop_level=' .. itemSplit[rest_offset]
    rest_offset = rest_offset + 1
  end

  -- Gems
  local gems = {}

  -- Get sockets for artifacts
  local itemType = select(9, GetItemInfo(itemLink))
  if ((itemType == "INVTYPE_WEAPONMAINHAND" and getSocketSlotID() == INVSLOT_MAINHAND) or (itemType == "INVTYPE_SHIELD" and getSocketSlotID() == INVSLOT_OFFHAND)) then
    for _, frame in ipairs(mod.frameSlots) do
      if frame.itemLink and frame.info.artifact_gem_id then
        local gemLink = useEquippedRelics and frame.equippedItem or frame.itemLink
        if gemLink then
          local gemDetail = string.match(gemLink, "item[%-?%d:]+")
          gems[#gems + 1] = string.match(gemDetail, "item:(%d+):" )
        elseif flags == 256 then
          gems[#gems + 1] = "0"
        end
      end
    end

  -- Sockets for non-artifacts
  else
    for i=1, 4 do -- hardcoded here to just grab all 4 sockets
      local _, gemLink = GetItemGem(itemLink, i)
      if gemLink then
        local gemDetail = string.match(gemLink, "item[%-?%d:]+")
        gems[#gems + 1] = string.match(gemDetail, "item:(%d+):" )
      elseif flags == 256 then
        gems[#gems + 1] = "0"
      end
    end
  end


  if #gems > 0 then
    simcItemOptions[#simcItemOptions + 1] = 'gem_id=' .. table.concat(gems, '/')
  end

  return table.concat(simcItemOptions, ',')
end

function mod:BuildOutput()
  local text = ""

  if mod.mode == "compare" then
    text = text .. "# Append this to your SimC profile to compare your current loadout against a proposed loadout\n"
    local profile, loadout = self:BuildLoadout("equipped", {})
    local profile2, _ = self:BuildLoadout("selected", loadout)
    text = text .. profile
    if #profile2 > 0 then
      text = text .. '\ncopy="' .. UnitName("player") .. ' - ' .. select(2, GetSpecializationInfo(GetSpecialization())).. ' (Comparison)"\n'
      text = text .. profile2
    end
  elseif mod.mode == "build" then
    text = text .. "# Paste into your SimC profile to sim this loadout\n"
    text = text .. self:BuildLoadout("selected", {})
  end
  mod.editBox:SetText(text .. "\n")
  mod.CopyFrame:Show()
end

function mod:BuildLoadout(source, existingItems)
  local text = ""
  local loadout = {}
  local talents = self:BuildTalentString() .. "\n"
  local artifact = (self:BuildArtifactString() or lastArtifactString or "").. "\n"
  if existingItems.talents ~= talents then
    text = text .. talents
  end
  if existingItems.artifact ~= artifact then
    text = text .. artifact
  end
  loadout.artifact = artifact
  loadout.talents = talents

  local finger, trinket, relicNames = 1, 1, {}
  for _, frame in ipairs(mod.frameSlots) do
    local name, itemLink, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice, classId, subClassId, link
    if source == "selected" and frame.itemLink then
      link = frame.itemLink
    elseif source == "equipped" then
      link = frame.equippedItem
    end

    if link then
      name, itemLink, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice, classId, subClassId = GetItemInfo(link)
    end
    local key = tostring(frame.info.slot_id) .. (name or "no item")

    if frame.info.slot_id and not frame.info.artifact_gem_id and link then
      local line = self:BuildSimcLine(itemLink, source == "equipped")
      if existingItems[key] ~= line then
        text = text .. frame.info.slot_name .. "=" .. line .. "\n"
        loadout[key] = line
      end
    elseif frame.info.artifact_gem_id then
      local existing
      if frame.equippedItem then
        existing = GetItemInfo(frame.equippedItem)
      end
      if name or existing then
        tinsert(relicNames, name or existing)
      end
    end
  end
  if #relicNames > 0 then
    local relics = "# Relics: " .. table.concat(relicNames, ', ') .. "\n"
    if existingItems.relics ~= relics then
      text =  text .. relics
    end
    loadout.relics = relics
  end
  return text, loadout
end